@@ -2,7 +2,7 @@
  *
  * idwcc: OAuth2/OIDC client program to test or validate OAuth2/OIDC AS
  *
- * Copyright 2021 Nicolas Mora <mail@babelouest.org>
+ * Copyright 2021-2022 Nicolas Mora <mail@babelouest.org>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU GENERAL PUBLIC LICENSE
@@ -51,7 +51,7 @@
 struct _callback_struct {
   struct _i_session * session;
   const char        * webapp_path;
-  uint                ciba_status;
+  unsigned int        ciba_status;
   pthread_mutex_t     change_lock;
   char                message[MESSAGE_MAX+1];
 };
